@@ -7,16 +7,21 @@ package Tickit::Widget::MenuBar;
 
 use strict;
 use warnings;
+use feature qw( switch );
 
 use base qw( Tickit::Widget::Menu::base );
 use Tickit::Style;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
+
+use Carp;
 
 use Tickit::RenderBuffer qw( LINE_SINGLE );
-use Tickit::Utils qw( textwidth );
 use List::Util qw( sum max );
 
+# Re-import the constant for compiletime use
+use constant separator => __PACKAGE__->separator;
+
 =head1 NAME
 
 C<Tickit::Widget::MenuBar> - display a menu horizontally
@@ -56,6 +61,9 @@ its name is never useful, and it should be added to a container widget, such
 as L<Tickit::Widget::VBox>, for longterm display. It does not have a C<popup>
 or C<dismiss> method.
 
+A single separator object can be added as an item, causing all the items after
+it to be right-justified.
+
 =head1 STYLE
 
 The default style pen is used as the widget pen. The following style pen 
@@ -69,13 +77,43 @@ The pen used to highlight the active menu selection
 
 =back
 
+The following style actions are used:
+
+=over 4
+
+=item highlight_next (<Right>)
+
+=item highlight_prev (<Left>)
+
+Highlight the next or previous item
+
+=item highlight_first (<F10>)
+
+Highlight the first menu item
+
+=item activate (<Enter>)
+
+Activate the highlighted item
+
+=item dismiss (<Escape>)
+
+Dismiss the menu
+
+=back
+
 =cut
 
 style_definition base =>
    rv => 1,
    highlight_rv => 0,
-   highlight_bg => "green";
+   highlight_bg => "green",
+   "<Right>" => "highlight_next",
+   "<Left>"  => "highlight_prev",
+   "<F10>"   => "highlight_first",
+   "<Enter>"  => "activate",
+   "<Escape>" => "dismiss";
 
+use constant KEYPRESSES_FROM_STYLE => 1;
 use constant WIDGET_PEN_FROM_STYLE => 1;
 
 sub lines
@@ -86,25 +124,45 @@ sub lines
 sub cols
 {
    my $self = shift;
-   return sum( map { textwidth $_->name } $self->items ) + 2 * ( $self->items - 1 );
+   return sum( map { $self->_itemwidth( $_ ) } 0 .. $self->items-1 ) + 2 * ( $self->items - 1 );
+}
+
+sub push_item
+{
+   my $self = shift;
+   my ( $item ) = @_;
+
+   if( $item == separator and grep { $_ == separator } $self->items ) {
+      croak "Cannot have more than one separator in a MenuBar";
+   }
+
+   $self->SUPER::push_item( $item );
 }
 
 sub reshape
 {
    my $self = shift;
 
-   $self->{itemcols} = \my @cols;
-   $self->{itemwidths} = \my @widths;
+   $self->{itempos} = \my @pos;
 
    my $items = $self->{items};
    my $col = 0;
+   my $separator_at;
    foreach my $idx ( 0 .. $#$items ) {
-      $cols[$idx] = $col;
-      $col += ( $widths[$idx] = textwidth $items->[$idx]->name );
+      $separator_at = $idx, next if $items->[$idx] == separator;
+
+      $pos[$idx] = [ $col, undef ];
+      $col += $self->_itemwidth( $idx );
+      $pos[$idx][1] = $col;
       $col += 2;
    }
 
-   push @cols, $col;
+   if( defined $separator_at ) {
+      $col -= 2; # undo
+      my $spare = $self->window->cols - $col;
+
+      $pos[$_][0] += $spare, $pos[$_][1] += $spare for $separator_at+1 .. $#$items;
+   }
 }
 
 sub pos2item
@@ -114,33 +172,32 @@ sub pos2item
 
    $line == 0 or return ();
 
-   my $cols   = $self->{itemcols};
-   my $widths = $self->{itemwidths};
+   my $items = $self->{items};
+   my $pos   = $self->{itempos};
 
-   foreach my $idx ( 0 .. $#$widths ) {
-      next unless $col < $cols->[$idx+1];
-      $col -= $cols->[$idx];
+   foreach my $idx ( 0 .. $#$items ) {
+      next if !defined $pos->[$idx]; # separator
+      last if     $col < $pos->[$idx][0];
+      next unless $col < $pos->[$idx][1];
 
-      return () if $col >= $widths->[$idx];
-      return ( $self->{items}->[$idx], $idx, $col );
+      $col -= $pos->[$idx][0];
+
+      return () if $col < 0;
+      return ( $items->[$idx], $idx, $col );
    }
 
    return ();
 }
 
-sub render_item
+sub redraw_item
 {
    my $self = shift;
-   my ( $idx, $rb ) = @_;
-
-   my $item = $self->{items}->[$idx];
-
-   $rb->goto( 0, $self->{itemcols}->[$idx] );
-
-   my $is_highlight = defined $self->{active_idx} && $idx == $self->{active_idx};
-   $rb->text( $item->name, $is_highlight ? $self->get_style_pen( "highlight" ) : undef );
-
-   $rb->erase( 2 );
+   my ( $idx ) = @_;
+   $self->window->expose( Tickit::Rect->new(
+      top => 0, lines => 1,
+      left  => $self->{itempos}[$idx][0],
+      right => $self->{itempos}[$idx][1],
+   ) );
 }
 
 sub render_to_rb
@@ -149,9 +206,28 @@ sub render_to_rb
    my ( $rb, $rect ) = @_;
 
    if( $rect->top == 0 ) {
+      $rb->goto( 0, 0 );
+
       my @items = $self->items;
       foreach my $idx ( 0 .. $#items ) {
-         $self->render_item( $idx, $rb );
+         my $item = $items[$idx];
+         next if $item == separator;
+
+         my ( $left, $right ) = @{ $self->{itempos}[$idx] };
+         last if $left > $rect->right;
+         next if $right < $rect->left;
+
+         $rb->erase_to( $left );
+
+         my $pen = defined $self->{active_idx} && $idx == $self->{active_idx}
+                     ? $self->get_style_pen( "highlight" ) : undef;
+
+         $rb->savepen;
+         $rb->setpen( $pen );
+
+         $item->render_label( $rb, $right - $left, $self );
+
+         $rb->restore;
       }
 
       $rb->erase_to( $rect->right );
@@ -169,7 +245,11 @@ sub popup_item
 
    my $items = $self->{items};
 
-   my $col = $self->{itemcols}->[$idx];
+   my $col = $self->{itempos}[$idx][0];
+
+   my $rightmost = $self->window->cols - $items->[$idx]->cols;
+   $col = $rightmost if $col > $rightmost;
+
    $items->[$idx]->popup( $self->window, 1, $col );
 }
 
@@ -188,6 +268,36 @@ sub dismiss
    $self->redraw;
 }
 
+sub on_key
+{
+   my $self = shift;
+
+   # Always eat all the keys as there's never anything higher to pass them to
+   return 1;
+}
+
+# MenuBar always expands on highlight
+sub key_highlight_next
+{
+   my $self = shift;
+   $self->SUPER::key_highlight_next;
+   $self->expand_item( $self->{active_idx} );
+}
+
+sub key_highlight_prev
+{
+   my $self = shift;
+   $self->SUPER::key_highlight_prev;
+   $self->expand_item( $self->{active_idx} );
+}
+
+sub key_highlight_first
+{
+   my $self = shift;
+   defined $self->{active_idx} or $self->expand_item( 0 );
+   return 1;
+}
+
 sub on_mouse_item
 {
    my $self = shift;
